divert(-1)

################################### Header ###################################

# "sage.m4"	K. J. Turner (kjt@cs.stir.ac.uk)	04/06/96

# This "m4" macro file contains templates for generating connection-oriented
# service specifications in LOTOS according to the SAGE (Service Attribute
# Generator) approach

# Copyright 1996 K. J. Turner, University of Stirling

############################### General Macros ###############################

# "line_max" defines the target line length

define(line_max,55)

# "ltrs_of(string)" produces the same string with digits deleted

define(ltrs_of,`translit($1,0123456789)'))

ltrs_of(Addr1|Addr2|Bool|Data)

# "lower_of(string)" produces the same string with upper case letters changed
# to lower case

define(lower_of,
  `translit(
    $1,ABCDEFGHIJKLMNOPQRSTUVWXYZ,abcdefghijklmnopqrstuvwxyz)'))

# "upper_of(string)" produces the same string with lower case letters changed
# to upper case

define(upper_of,
  `translit(
    $1,abcdefghijklmnopqrstuvwxyz,ABCDEFGHIJKLMNOPQRSTUVWXYZ)'))

# "get_word(string,num)" gets the space-terminated word from "string" with
# number "num" (0 is first word), returning a null string if there is no such
# word

define(get_word,
  `define(`ind',`index($1,` ')')ifelse(ind,-1,,
    `ifelse(eval($2 < 0),1,,
      `ifelse(eval($2 = 0),1,
        `substr($1,0,ind)',
	  `get_word(substr($1,incr(ind)),eval($2 - 1))')')')')

# "intrn_of(string)" produces the same string with function calls
# changed into internal form, i.e. "f(a,b)" translated to
# "f[a|b]". This is because "(", "," and ")" are special to m4.
# Newlines, blanks and tabs are also removed.

define(intrn_of,`translit(translit(`$1',`(,)',[|]),`
' 	)')

# "extrn_of(string)" produces the external equivalent of the
# internal form of function calls

define(extrn_of,`translit($1,[|],`(,)')')

# "map_funs(string,macro)" calls the macro for each function call
# in the string. This macro is called with the function name as
# its first parameter, and the function parameter list as its
# second parameter.
#
# The macro uses the macros "mf_ind1" and "mf_ind2" which must
# not be altered by the argument macro.

define(map_funs,
  `define(`mf_ind1',
    index($1,[))define(
      `mf_ind2',index($1,]))ifelse(
	mf_ind1,-1,,
	  mf_ind2,-1,,
	     `$2(
	       substr($1,0,mf_ind1),substr($1,incr(mf_ind1),
		   eval(mf_ind2-mf_ind1-1)))'``''`map_funs(
		     substr($1,incr(mf_ind2)),`$2')')')

# "map_args(string,macro)" calls the macro for each parameter in
# the string. The macro is called with the function parameter list
# as its only parameter.
#
# The macro uses the macros "ma_ind1" which must not be altered
# by the argument macro.

define(map_args,
  `define(`ma_ind1',
    index($1,|))ifelse(
      $1,,,
	ma_ind1,-1,`$2($1)',
	  `$2(substr($1,0,ma_ind1))'``''`map_args(
	    substr($1,incr(ma_ind1)),`$2')')')

# "lab_1" and "lab_2" produce the labels to distinguish the parameters of two
# constructor operations. The labels must be alphabetic.

define(lab_1,A)

define(lab_2,B)

# "lab_args(string,label)" labels each argument in the string by
# appending the given label

define(lab_args,
  `undefine(`list1')define(`lab_arg',
    `define(`par',
      $'1`)ifdef(`list1',
        `define(`list1',list1|par`'$2)',
          `define(`list1',par`'$2)')')map_args($1,
            `lab_arg')list1')

# "file_delete(file)" deletes file "file"

define(file_delete,`syscmd(rm $1)')

# "file_save(file,text)" saves the definition of a process "text" in file
# "file"

define(file_save,`syscmd(cat >> $1 <<EOF
$2
EOF)')

# "igid_pair(dir)" produces an igid pair for direction "dir" ("12" or "21")

define(igid_pair,
 `ifelse($1,12,`igid1, igid2',`igid2, igid1')')

# "igid_swap(beh)" produces a behaviour corresponding to "beh" but with igids
# swapped

define(igid_swap,
  `define(`igid1',
    ``igid2'')define(`igid2',
      ``igid1'')$1`'undefine(`igid1')undefine(`igid2')')

# "proc_of(string)" produces the name in a process or procedure instantiation
# such as "Conn" in "Conn(alpha,beta)" or "Conn[gamma,delta]"

define(proc_of,
  `ifelse(len($1),0,,
    `define(`char',substr(`$1',0,1))ifelse(char,` ',,
      `ifelse(char,`(',,
        `ifelse(char,`[',,
	  `char`'proc_of(substr(`$1',1))')')')')')

# "pars_of(string)" produces the parameters in a process or procedure
# instantiation such as "(alpha,beta)" in "Conn(alpha,beta)" or
# "[gamma,delta]" in "Conn[gamma,delta]"

define(pars_of,
  `define(`ind',
    index($1,`('))ifelse(ind,-1,`define(`ind',
      index($1,`['))ifelse(ind,-1,,substr($1,ind))',substr($1,ind))')

# "group_of(string)" produces the group in a process or procedure
# instantiation such as "Conn" in "ConnReq(alpha,beta)" or
# "ConnReq[gamma,delta]" by removing the last three characters of the name

define(group_of,
  `define(`proc',`proc_of($1)')substr(proc,0,eval(len(proc)-3))')

# "gate_qual(num,beh)" produces a behaviour corresponding to "beh" but with
# number "num" appended to the gate name

define(gate_qual,
  `define(`ind',
    index($2,gate))ifelse(ind,-1,$2,
      `define(`ind',eval(ind+len(gate)))substr($2,0,ind)()$1substr($2,ind)')')

###################### Basic Primitive Type Macros #########################

# "lib_type" invokes library type definitions

define(lib_type,`
  library
    Boolean, NaturalNumber, OctetString, String, Set
  endlib')

# "igid_type" produces an interaction group identifier type definition

define(igid_type,`
  type IGId is NaturalNumber renamedby
    sortnames IGId for Nat
  endtype (* IGId *)')

# "igid_set_type" produces an interaction group identifier set type definition

define(igid_set_type,`
  type IGIdSet is Set actualizedby IGId, Boolean using
    sortnames
      IGIdSet	for Set
      IGId	for Element
      Bool	for FBool
  endtype (* IGId *)')

# "addr_type" produces an address type definition

define(addr_type,`
  type Addr is NaturalNumber renamedby
    sortnames Addr for Nat
  endtype (* Addr *)')

# "data_type" produces a data type definition

define(data_type,`
  type Data is OctetString renamedby
    sortnames Data for OctetString
  endtype (* Data *)')

# "orig_type" produces an originator type definition

define(orig_type,`
  type Orig is Boolean, NaturalNumber
    sorts Orig
    opns
      User, Prov, Other	:		-> Orig
      Ord		: Orig		-> Nat
      _ eq _, _ ne _	: Orig, Orig	-> Bool
    eqns
      forall orig`'lab_1, orig`'lab_2 : Orig
        ofsort Nat
	  Ord (User)	= 0;
	  Ord (Prov)	= Succ (Ord (User));
	  Ord (Other)	= Succ (Ord (Prov));
        ofsort Bool
	  orig`'lab_1 eq orig`'lab_2 = Ord (orig`'lab_1) eq Ord (orig`'lab_2);
	  orig`'lab_1 ne orig`'lab_2 = not (orig`'lab_1 eq orig`'lab_2);
  endtype (* Orig *)')

# "reas_type" produces a reason type definition

define(reas_type,`
  type Reas is Boolean, NaturalNumber
    sorts Reas
    opns
      UserInit, ProvInit, Unknown	:		-> Reas
      Ord				: Reas		-> Nat
      _ eq _, _ ne _			: Reas, Reas	-> Bool
    eqns
      forall reas`'lab_1, reas`'lab_2 : Reas
        ofsort Nat
	  Ord (UserInit)	= 0;
	  Ord (ProvInit)	= Succ (Ord (UserInit));
	  Ord (Unknown)		= Succ (Ord (ProvInit));
        ofsort Bool
	  reas`'lab_1 eq reas`'lab_2 = Ord (reas`'lab_1) eq Ord (reas`'lab_2);
	  reas`'lab_1 ne reas`'lab_2 = not (reas`'lab_1 eq reas`'lab_2);
  endtype (* Reas *)')

# "kind_type" produces a service primitive kind type definition

define(kind_type,`
  type Kind is NaturalNumber, Boolean
    sorts Kind
    opns
      Ord		: Kind		-> Nat
      _ eq _, _ ne _	: Kind, Kind	-> Bool
    eqns
      forall kind`'lab_1, kind`'lab_2 : Kind
        ofsort Bool
	  kind`'lab_1 eq kind`'lab_2 = Ord (kind`'lab_1) eq Ord (kind`'lab_2);
	  kind`'lab_1 ne kind`'lab_2 = Ord (kind`'lab_1) ne Ord (kind`'lab_2);
  endtype (* Kind *)')

# "group_type" produces a service primitive group type definition

define(group_type,`
  type Group is NaturalNumber, Boolean
    sorts Group
    opns
      Ord		: Group		-> Nat
      _ eq _, _ ne _	: Group, Group	-> Bool
    eqns
      forall group`'lab_1, group`'lab_2 : Group
        ofsort Bool
	  group`'lab_1 eq group`'lab_2 = Ord (group`'lab_1) eq Ord (group`'lab_2);
	  group`'lab_1 ne group`'lab_2 = Ord (group`'lab_1) ne Ord (group`'lab_2);
  endtype (* Group *)')

###################### Primitive Queue Type Macros #########################

# "primaux_type" produces a service primitive auxiliary type definition

define(primaux_type,`
  type PrimAux is Kind, Group
    sorts Prim
    opns
      KindOf		: Prim		-> Kind
      GroupOf		: Prim		-> Group
      IsReq, IsInd	: Prim		-> Bool
      IsRes, IsCon	: Prim		-> Bool
      IsKind		: Prim, Kind	-> Bool
      IsGroup		: Prim, Group	-> Bool
      IsNextPrim	: Prim, Prim	-> Bool
      _ eq _, _ ne _	: Prim, Prim	-> Bool
    eqns
      forall prim : Prim, kind : Kind, group : Group
        ofsort Bool
	  IsKind (prim, kind)	= KindOf (prim) eq kind;
	  IsGroup (prim, group)	= GroupOf (prim) eq group;
  endtype (* PrimAux *)')

# "primqaux_type" produces a service primitive queue auxiliary type
# definition

define(primqaux_type,`
  type PrimQueueAux is String actualizedby Prim using
    sortnames
      Prim for Element
      Bool for FBool
      PrimQueue for String
    opnnames
      <<>> for <>
      PrimQueue for String
  endtype (* PrimQueueAux *)')

# "primq_type" produces a service primitive queue type definition

define(primq_type,`primqaux_type

  type PrimQueue is PrimQueueAux
    opns
      Head		: PrimQueue		-> Prim
      Tail		: PrimQueue		-> PrimQueue
      RemPrev		: Prim, PrimQueue	-> PrimQueue
      RepPrev		: Prim, PrimQueue	-> PrimQueue
      IsNextQueue	: Prim, PrimQueue	-> Bool
      IsAheadQueue	: Prim, PrimQueue	-> Bool
    eqns
      forall prim, prim`'lab_1, prim`'lab_2 : Prim, primq : PrimQueue
        ofsort Prim
	  Head (prim`'lab_2 + <<>>)			= prim`'lab_2;
	  Head (prim`'lab_2 + (prim`'lab_1 + primq))	= Head (prim`'lab_1 + primq);
	ofsort PrimQueue
	  Tail (prim`'lab_2 + <<>>)			= <<>>;
	  Tail (prim`'lab_2 + (prim`'lab_1 + primq))	= prim`'lab_2 + Tail (prim`'lab_1 + primq);
	  RemPrev (prim`'lab_2, <<>>)			= <<>>;
	  IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq)) =>
	    RemPrev (prim`'lab_2, prim`'lab_1 + primq)	= Tail (prim`'lab_1 + primq);
	  not (IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq))) =>
	    RemPrev (prim`'lab_2, prim`'lab_1 + primq)	=
	      RemPrev (prim`'lab_2, Tail (prim`'lab_1 + primq)) + Head (prim`'lab_1 + primq);
	  RepPrev (prim`'lab_2, <<>>)			= <<>>;
	  IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq)) =>
	    RepPrev (prim`'lab_2, prim`'lab_1 + primq)	= Tail (prim`'lab_1 + primq) + prim`'lab_2;
	  not (IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq))) =>
	    RepPrev (prim`'lab_2, prim`'lab_1 + primq)	=
	      RepPrev (prim`'lab_2, Tail (prim`'lab_1 + primq)) + Head (prim`'lab_1 + primq);
        ofsort Bool
	  IsNextQueue (prim`'lab_2, <<>>)		= false;
	  IsNextQueue (prim`'lab_2, prim`'lab_1 + primq)	=
	    IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq));
	  IsAheadQueue (prim`'lab_2, <<>>)		= false;
	  IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq)) =>
	    IsAheadQueue (prim`'lab_2, prim`'lab_1 + primq)	= true;
	  not (IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq))),
	   GroupOf (prim`'lab_2) eq GroupOf (Head (prim`'lab_1 + primq)) =>
	    IsAheadQueue (prim`'lab_2, prim`'lab_1 + primq)	= false;
	  not (IsNextPrim (prim`'lab_2, Head (prim`'lab_1 + primq))),
	   GroupOf (prim`'lab_2) ne GroupOf (Head (prim`'lab_1 + primq)) =>
	    IsAheadQueue (prim`'lab_2, prim`'lab_1 + primq)	=
	      IsAheadQueue (prim`'lab_2, Tail (prim`'lab_1 + primq));
  endtype (* PrimQueue *)')

###################### Primitive Type Macros #########################

# "prim_con_opn(operation,parameters)" produces the signature of
# a constructor operation with given name and parameters

define(prim_con_opn,
  `      $1 : ifelse($2,,,`extrn_of(ltrs_of($2)) ')-> Prim
')

# "prim_con_opns(string)" produces the signatures of the
# constructor operations given as functions in the string

define(prim_con_opns,`map_funs($1,`prim_con_opn')')

# "prim_kind_opn(operation,parameters)" produces the signature of the kind
# operation for the primitive with given name and parameters

define(prim_kind_opn,
  `ifdef(`list1',
    `define(`line_len',
      eval(line_len + len(|$1Kind)))ifelse(eval(line_len < line_max),1,
        `define(`list1',list1|$1Kind)',`define(`list1',list1|
        $1Kind)define(`line_len',0)')',`define(`list1',$1Kind)')')

# "prim_kind_opns(string)" produces the signature of the
# kind operations for primitives given as functions in the string

define(prim_kind_opns,
  `undefine(`list1')define(`line_len',0)map_funs(
    $1,`prim_kind_opn')      extrn_of(list1) : -> Kind')

# "prim_group_opn(operation,parameters)" produces the signature of the group
# operation for the primitive with given name and parameters.

define(prim_group_opn,
  `define(`group1',`group_of($1)')ifdef(`list1',
    `define(`ind1',index(list1,group1))ifelse(ind1,-1,`define(`line_len',
      eval(line_len + len(|group1)))ifelse(eval(line_len < line_max),1,
        `define(`list1',list1|group1)',`define(`list1',list1|
        $1Kind)define(`line_len',0)')')',`define(`list1',group1)')')

# "prim_group_opns(string)" produces the signature of the
# group operations for primitives given as functions in the string

define(prim_group_opns,
  `undefine(`list1')define(`line_len',0)map_funs(
    $1,`prim_group_opn')      extrn_of(list1) : -> Group')

# "prim_opns(string)" produces the signatures of the operations
# for the functions in the string, as well as the auxiliary
# operations

define(prim_opns,
`prim_con_opns($1)`'prim_kind_opns($1)
prim_group_opns($1)')

# "prim_var_dec(parameter)" produces the global variable
# declarations for the given parameter of a constructor operation.
# Declarations of the same sort are grouped in an order such as:
#
#  Par,ParA,ParB,Par1,Par1A,Par1B,Par2,Par2A,Par2B : Par

define(prim_var_dec,
  `ifdef(`list1',
    `ifelse(index(list1,`$1|'),-1,
      `ifelse(index(list1,`$1 '),-1,
        `define(`ind1',
          index(list1,ltrs_of($1)))ifelse(ind1,-1,
          `define(`list1',list1|
        $1|$1`'lab_1|$1`'lab_2 : ltrs_of($1))',
            `ifelse($1,ltrs_of($1),
              `define(`list1',
                substr(list1,0,
                  ind1)$1|$1`'lab_1|$1`'lab_2|substr(list1,
                    ind1))',
                `define(`ind2',
                  `index(substr(list1,ind1),` ')')define(`ind2',
                    eval(ind1+ind2))define(`list1',
                      substr(list1,0,
                        ind2)|$1|$1`'lab_1|$1`'lab_2`'substr(list1,
                          ind2))')')')')',
      `define(`list1',`
        $1|$1`'lab_1|$1`'lab_2 : ltrs_of($1)')')')

# "prim_for_dec(operation,parameters)" produces the global
# variable declarations for the primitive with given parameters

define(prim_for_dec,`map_args(lower_of($2),`prim_var_dec')')

# "prim_for_decs(string)" produces the global variable declarations for the
# primitives given as functions given in the string

define(prim_for_decs,
  `undefine(`list1')map_funs([prim]$1,`prim_for_dec')extrn_of(`list1')')

# "prim_kind_eqn(operation,parameters)" produces the equations for the kind
# operation of the primitive with the given name and parameters

define(prim_kind_eqn,
  `          KindOf ($1`'ifelse($2,,,
    ` (extrn_of(lower_of($2)))'))	= proc_of($1)Kind;
')

# "prim_kind_eqns(string)" produces the equations for the kind operations of
# primitives given as functions in the string

define(prim_kind_eqns,`map_funs($1,`prim_kind_eqn')')

# "prim_kind_ord_eqn(operation,parameters)" produces the equation for the
# ordinal value of the kind operation defined by the primitive with the
# given name and parameters

define(prim_kind_ord_eqn,
  `          Ord ($1Kind)	= ord;define(`ord',
      Succ (Ord ($1Kind)))
')

# "prim_kind_ord_eqns(string)" produces the equations for the ordinal
# values of kind operations defined by functions in the string

define(prim_kind_ord_eqns,`define(`ord',0)map_funs($1,`prim_kind_ord_eqn')')

# "prim_group_eqn(operation,parameters)" produces the equations for the group
# operation of the primitive with the given name and parameters

define(prim_group_eqn,
  `          GroupOf `('$1`'ifelse($2,,,
    ` `('extrn_of(lower_of($2))`)'')`)'	= group_of($1);
')

# "prim_group_eqns(string)" produces the equations for the group operations of
# primitives given as functions in the string

define(prim_group_eqns,`map_funs($1,`prim_group_eqn')')

# "prim_group_ord_eqn(operation,parameters)" produces the equation for the
# ordinal value of the group operation defined by the primitive with the
# given name and parameters

define(prim_group_ord_eqn,
  `define(`eqn1',)define(`group1',group_of($1))ifelse(`list1',,,
    `define(`ind1',index(list1,group1))ifelse(ind1,-1,
      `define(`eqn1',`          Ord (group1)	= ord;
define(`ord',Succ (Ord (group1)))')')')define(`list1',list1`'eqn1)')

# "prim_group_ord_eqns(string)" produces the equations for the ordinal
# values of group operations defined by functions in the string

define(prim_group_ord_eqns,
  `define(`list1',)define(`ord',0)map_funs($1,`prim_group_ord_eqn')list1')

# "prim_eq_arg(parameter)" produces a conditional requiring
# equality of each parameter in the string

define(prim_eq_arg,
  `ifdef(`list1',
    `define(`list1',`('list1`)' and `('$1`'lab_1 eq $1`'lab_2`)')',
      `define(`list1',$1`'lab_1 eq $1`'lab_2)')')

# "prim_eq_args(string)" produces a compound conditional requiring
# equality of each argument in the string

define(prim_eq_args,
  `undefine(`list1')map_args(
    lower_of($1),`prim_eq_arg')            list1;')

# "prim_eq_eqn(operation,parameters)" produces the equation for
# the equality operation corresponding to the given operation and
# parameters

define(prim_eq_eqn,
  `          ifelse($2,,$1 eq $1 = true;,
    $1 (extrn_of(lab_args(lower_of($2),
      `lab_1'))) eq $1 (extrn_of(lab_args(lower_of($2),
        `lab_2'))) =
prim_eq_args($2))
')

# "prim_eq_eqns(string)" produces the equations for the equality
# operation corresponding to the functions in the string

define(prim_eq_eqns,`          KindOf (prim`'lab_1) ne KindOf (prim`'lab_2) =>
            prim`'lab_1 eq prim`'lab_2 = false;
map_funs($1,
  `prim_eq_eqn')          prim`'lab_1 ne prim`'lab_2 = not (prim`'lab_1 eq prim`'lab_2);
')

# "prim_dir_eqn(operation,parameters)" produces the equation for
# the recogniser operation corresponding to the given operation
# and parameters. The global "dir1" defines the direction (Req/Ind/Res/Con).

define(prim_dir_eqn,
  `ifelse(eval(index($1,dir1) != -1),1,`ifdef(`list1',
    `define(`line_len',
      eval(line_len + len(` or IsKind (Prim, $1Kind)')))ifelse(eval(line_len<line_max),1,
        `define(`list1',
	  list1 or IsKind (Prim, $1Kind))',`define(`list1',
	    list1 or
              IsKind (Prim, $1Kind))define(`line_len',0)')',
	  `define(`list1',`            IsKind (Prim, $1Kind)')')')')

# "prim_dir_eqns(string,direction)" produces the equations for the recogniser
# operations corresponding to the functions in the string

define(prim_dir_eqns,
  `undefine(`list1')define(`line_len',0)define(`dir1',
    $2)map_funs($1,`prim_dir_eqn')ifdef(`list1',
      `          Is`'dir1 (Prim) =
list1;
')')

# "prim_rec_eqns(string)" produces the equations for the recogniser
# operations corresponding to the functions in the string

define(prim_rec_eqns,
  `prim_dir_eqns($1,
    Req)prim_dir_eqns($1,
      Ind)prim_dir_eqns($1,
        Res)prim_dir_eqns($1,
          Con)')

# "prim_not_next_eqn(operation,parameters)" produces the equation for the
# not "next primitive" operation corresponding to the current and
# previous operation and parameters ("oper2" and "oper1" respectively). The
# comma in the list is "protected" by making it an internal "|"

define(prim_not_next_eqn,
  `define(`oper2',$1)ifdef(`oper1',
    `ifelse(group_of(oper2),group_of(oper1),
      `ifdef(`list1',`define(`list1',list1|
            not (IsKind (prim`'lab_1, oper2`'Kind) and IsKind (prim`'lab_2, oper1`'Kind)))',
        `define(`list1',
not (IsKind (prim`'lab_1, oper2`'Kind) and IsKind (prim`'lab_2, oper1`'Kind)))')')')define(`oper1',
          oper2)')

# "prim_not_next_eqns(string)" produces the equations for the not
# "next primitive" operation corresponding to the functions in the string

define(prim_not_next_eqns,
  `undefine(`list1')undefine(`oper1')map_funs($1,
    `prim_not_next_eqn')ifdef(`list1',`          extrn_of(list1) =>
              IsNextPrim (prim`'lab_1, prim`'lab_2) = false;')')

# "prim_is_next_eqn(operation,parameters)" produces the equation for
# the "next primitive" operation corresponding to the current and
# previous operation and parameters ("oper2"/"pars2" and "oper1"/"pars1"
# respectively)

define(prim_is_next_eqn,
  `define(`oper2',$1)define(`pars2',$2)ifdef(`oper1',
    `ifelse(group_of(oper2),group_of(oper1),
      `          IsNextPrim (oper2`'ifelse(pars2,,,
        ` (extrn_of(lab_args(lower_of(pars2),
        `lab_1')))'), oper1`'ifelse(pars1,,,
        ` (extrn_of(lab_args(lower_of(pars1),
        `lab_2')))')) = ifelse(pars1,pars2,ifelse(pars1,,true;,`
prim_eq_args($2)'),true;)
')')define(`oper1',oper2)define(`pars1',pars2)')

# "prim_is_next_eqns(string)" produces the equations for the "next primitive"
# operation corresponding to the functions in the string

define(prim_is_next_eqns,
  `undefine(`oper1')undefine(`pars1')map_funs($1,`prim_is_next_eqn')')

# "prim_next_eqns(string)" produces the equations for the "next primitive"
# operation corresponding to the functions in the string

define(prim_next_eqns,
  `prim_not_next_eqns($1)
prim_is_next_eqns($1)')

# "prim_eqns(string)" produces the signatures of the operations
# for the functions in the string, as well as the auxiliary
# operations

define(prim_eqns,`        ofsort Kind
prim_kind_eqns($1)        ofsort Nat
prim_kind_ord_eqns($1)        ofsort Group
prim_group_eqns($1)        ofsort Nat
prim_group_ord_eqns($1)        ofsort Bool
prim_eq_eqns($1)        ofsort Bool
prim_rec_eqns($1)        ofsort Bool
prim_next_eqns($1)')

# "prim_type(string)" produces a service primitive type definition based
# on the primitives given as functions in the string

define(prim_type,`primaux_type

  type Prim is Addr, Data, Reas, PrimAux
    opns
prim_opns($1)
    eqns
      forall`'prim_for_decs($1)
prim_eqns($1)  endtype (* Prim *)')

##################### All Type Definitions #######################

# "all_types(string)" produces all type definitions based on the primitives
# given as functions in the string

define(all_types,`lib_type
igid_type
igid_set_type
addr_type
data_type
orig_type
reas_type
kind_type
group_type
prim_type($1)
primq_type
')

# "add_prim_types(pattern,group1,group2)" adds the service primitive names and
# parameters for tries "group1" and (if relevant) acknowledgements "group2";
# the pattern controls which request/indication/response/confirm primitives
# are generated

define(add_prim_types,
 `define(`prims1',)define(`proc1',
  `proc_of($2)')define(`pars1',
   `intrn_of(pars_of($2))')define(`proc2',
    `proc_of($3)')define(`pars2',
     `intrn_of(pars_of($3))')ifelse($1,provider_initiated,
      `ifelse(index(prims,proc1),-1,`define(`prims1',proc1`'Ind`'pars1)')',
       $1,unconfirmed,`ifelse(index(prims,proc1),-1,`define(`prims1',
        proc1`'Req`'pars1`'proc1`'Ind`'pars1)')',
         $1,provider_confirmed,`ifelse(index(prims,proc1),-1,`define(`prims1',
          proc1`'Req`'pars1`'proc1`'Ind`'pars1`'ifelse(proc1,proc2,
          proc1`'Con`'pars2))')ifelse(index(prims,proc2),
	  -1,`ifelse(proc1,proc2,,`define(`prims1',
	  prims1`'proc2`'Ind`'pars2)')')',
	   $1,user_provider_confirmed,`ifelse(index(prims,proc1),-1,
	    `define(`prims1',
	    proc1`'Req`'pars1`'proc1`'Ind`'pars1`'ifelse(proc1,proc2,
	    proc1`'Res`'pars2`'proc1`'Con`'pars2))')ifelse(index(prims,
	    proc2),-1,`ifelse(proc1,proc2,,`define(`prims1',
	    prims1`'proc2`'Req`'pars2`'proc2`'Ind`'pars2)')')',
	     $1,user_confirmed,`ifelse(index(prims,proc1),-1,
	      `define(`prims1',
	      proc1`'Req`'pars1`'proc1`'Ind`'pars1`'ifelse(proc1,proc2,
	      proc1`'Res`'pars2`'proc1`'Con`'pars2))')ifelse(index(prims,
	      proc2),-1,`ifelse(proc1,proc2,,`define(`prims1',
	      prims1`'proc2`'Req`'pars2`'proc2`'Ind`'pars2)')')',
	       `define(`prims1',
	       )errprint(`Unknown pattern $1!')')ifelse(`prims1',,,
	        `define(`prims',prims`'prims1)')')

# "prims" gives the groups and parameters of service facilities

define(prims,)

############################# Facility Processes #############################

# "add_facil_attrs(proc,dir,patt,prop)" adds the facility process name,
# direction, pattern and property, preceded by a `:' and followed by a space

define(add_facil_attrs,
  `define(`facil_attrs',:$1 $2 $3 $4 `'facil_attrs)')

# "facil_file" gives the temporary filename used for facility processes

define(facil_file,maketemp(/tmp/facilXXXXXX))

# "dir_of(beh)" returns the direction of "beh" if it is a facility, else a
# null string

define(dir_of,
  `define(`ind',`index(facil_attrs,:proc_of($1) )')ifelse(ind,-1,,
    `get_word(substr(facil_attrs,ind),1)')')

# "facil_procno" gives the next number to be used for a facility process

define(facil_procno,0)

# "facility(dir,patt,prop,group1,group2)" produces a process definition and
# process instantiation in direction "dir" for pattern "patt"
# (provider_initiated, unconfirmed, provider_confirmed,
# user_provider_confirmed, user_confirmed), property "prop" (single,
# consecutive, ordered, reliable, unreliable); the service primitive group and
# parameters for a try are given by "group1", while those for an
# acknowledgement (if relevant) are given by optional "group2"

define(facility,
  `define(`facil_procno',
    incr(facil_procno))define(`group1',proc_of($4))define(`procname1',
      group1`'Facil`'facil_procno)define(`procname2',
        procname1`'Aux)define(`procname3',
          procname1`'AuxA)add_facil_attrs(`procname1',
	    $1,$2,$3)add_prim_types($2,$4,$5)file_save(facil_file,
`  '(* group in direction $1 is $2 and $3 *)
  
  process procname1 [gate] (igid1, igid2 : IGId) : exit :=
 ifelse($3,single,
`   procname2 [gate] (igid1, igid2)',
    ifelse($3,consecutive,
`   procname2 [gate] (igid1, igid2)
  >>
    procname1 [gate] (igid1, igid2)',
 ifelse($3,ordered,
`   (````''''
      procname2 [gate] (igid1, igid2)
    |||
      procname1 [gate] (igid1, igid2)
    )
  ||
    procname3 [gate] (igid1, igid2, <<>>)',
 ifelse($3,reliable,
`   procname2 [gate] (igid1, igid2)
  |||
    procname1 [gate] (igid1, igid2)',
 ifelse($3,unreliable,
`   (``````''''''
      procname2 [gate] (igid1, igid2)
    [>
      i;		(* potential loss of messages *)
      exit
    )
  |||
    procname1 [gate] (igid1, igid2)',
`   stop		(*** "$3" is an unknown property ***)')))))
      
  where
facility_aux($2,group1)
ifelse($3,ordered,facility_ord($2,group1))
  endproc (* procname1 *)
)procname1 [gate] (igid_pair($1))')

# "facility_aux(patt,group)" produces a process definition for the
# behaviour in direction "dir" of facility pattern "patt" (provider_initiated,
# unconfirmed, provider_confirmed, user_provider_confirmed, user_confirmed),
# and service primitives "group"

define(facility_aux,
  `define(`procname',
      $2Facil`'facil_procno`'Aux)
    process procname [gate] (igid1, igid2 : IGId) : exit :=
 ifelse($1,provider_initiated,
`     gate ! igid2 ? prim1 : Prim [IsInd (prim1)];
      exit',
`     gate ! igid1 ? prim1 : Prim [IsReq (prim1) and IsGroup (prim1, $2)];
      gate ! igid2 ? prim2 : Prim [IsNextPrim (prim2, prim1)];'
 ifelse($1,unconfirmed,
`     exit',
 ifelse($1,provider_confirmed,
`     gate ! igid1 ? prim3 : Prim [IsNextPrim (prim3, prim2)];
      exit',
 ifelse($1,user_provider_confirmed,
`     (`````'''''
        gate ! igid2 ? prim3 : Prim [IsNextPrim (prim3, prim2)];
        gate ! igid1 ? prim4 : Prim [IsNextPrim (prim4, prim3)];
        exit
      []
        gate ! igid1 ? prim4 : Prim [IsNextPrim (prim4, prim2)];
        gate ! igid2 ? prim3 : Prim [IsNextPrim (prim3, prim2)];
        exit
      )',
 ifelse($1,user_confirmed,
`     gate ! igid2 ? prim3 : Prim [IsNextPrim (prim3, prim2)];
      gate ! igid1 ? prim4 : Prim [IsNextPrim (prim4, prim3)];
      exit',
`     stop		(*** "$1" is an unknown pattern ***)')))))
    endproc (* procname *)')

# "facility_ord(patt,group)" produces a process definition for ordering a
# facility in direction "dir" of pattern "patt" (provider_initiated,
# unconfirmed, provider_confirmed, user_confirmed), and service primitives
# "group"

define(facility_ord,
  `define(`procname',
      $2Facil`'facil_procno`'AuxA)
    process procname [gate]
     (igid1, igid2 : IGId, primq : PrimQueue) : noexit :=
 ifelse($1,provider_initiated,
`     gate ? igid ? prim : Prim;
      procname [gate] (igid1, igid2, primq)',
`     gate ! igid1 ? prim1 : Prim [IsReq (prim1) and IsGroup (prim1, $2)];
      procname [gate] (igid1, igid2, prim1 + primq)
    []
      gate ! igid2 ? prim2 : Prim [IsNextQueue (prim2, primq)];'
ifelse($1,unconfirmed,
`      procname [gate] (igid1, igid2, RemPrev (prim2, primq))',
 ifelse($1,provider_confirmed,
`      procname [gate] (igid1, igid2, RepPrev (prim2, primq))
    []
      gate ! igid1 ? prim3 : Prim [IsNextQueue (prim3, primq)];
      procname [gate] (igid1, igid2, RemPrev (prim3, primq))',
 ifelse($1,user_provider_confirmed,
`      procname [gate] (igid1, igid2, RepPrev (prim2, primq))
    []
      gate ! igid2 ? prim3 : Prim [IsNextQueue (prim3, primq)];
      procname [gate] (igid1, igid2, RepPrev (prim3, primq))
    []
      gate ! igid1 ? prim4 : Prim [IsNextQueue (prim4, primq)];
      procname [gate] (igid1, igid2, RemPrev (prim4, primq))',
 ifelse($1,user_confirmed,
`      procname [gate] (igid1, igid2, RepPrev (prim2, primq))
    []
      gate ! igid2 ? prim3 : Prim [IsNextQueue (prim3, primq)];
      procname [gate] (igid1, igid2, RepPrev (prim3, primq))
    []
      gate ! igid1 ? prim4 : Prim [IsNextQueue (prim4, primq)];
      procname [gate] (igid1, igid2, RemPrev (prim4, primq))',
`      stop		(*** "$1" is an unknown pattern ***)')))))
    endproc (* procname *)'
)

# "is_facil(beh)" checks if behaviour "beh" is a facility

define(is_facil,
  `define(`ind',`index(facil_attrs,:proc_of($1) )')eval(ind != -1)')

# "patt_of(beh)" returns the pattern of "beh" if it is a facility, else a null
# string

define(patt_of,
  `define(`ind',`index(facil_attrs,:proc_of($1) )')ifelse(ind,-1,,
    `get_word(substr(facil_attrs,ind),2)')')

# "prop_of(beh)" returns the property of "beh" if it is a facility, else a
# null string

define(prop_of,
  `define(`ind',`index(facil_attrs,:proc_of($1) )')ifelse(ind,-1,,
    `get_word(substr(facil_attrs,ind),3)')')

############################ Combination Processes ###########################

# "alternate(beh)" produces a process definition and process instantiation
# for behaviour "beh" alternating in each direction

define(alternate,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Alternate`'comb_procno)file_save(comb_file,
`  '(* proc_of($1) alternates in each direction *)
  
  process procname [gate] (igid1, igid2 : IGId) : noexit :=
    $1
  >>
    igid_swap($1)
  >>
    procname (igid1, igid2)
  endproc (* procname *)
)procname [gate] (igid1, igid2)')

# "colliding(facil)" produces a process definition and process instantiation
# for facility "facil" (which cannot be a combination) that may collide with
# itself; the following requires "facil" to be unconfirmed, but should
# take account of different patterns

define(colliding,
  `define(`comb_procno',
    incr(comb_procno))define(`procname1',
      Colliding`'comb_procno)define(`procname2',
        procname1`'Aux)define(`procname3',
          procname1`'AuxA)file_save(comb_file,
`  '(* proc_of($1) may collide with itself *)
  
  process procname1 [gate] (igid1, igid2 : IGId) : exit :=
ifelse(is_facil($1),0,
`    stop		(*** "proc_of($1)" is not one of the facilities ***)
  endproc (* procname1 *)
',
`ifelse(patt_of($1),unconfirmed,
`    hide gate()1``,'' gate()2 in
      (``''
	gate_qual(1,$1)
      |||
	igid_swap(gate_qual(2,$1))
      )
    |[gate()1``,'' gate()2]|
      procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    
  where
  
    process procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : exit :=
      (``''
	procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      []
	procname3 [gate``,'' gate()2``,'' gate()1] (igid1, igid2)
      )
    >>
      (
        procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      []
        exit
      )
    endproc (* procname2 *)
    
    process procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : exit :=
      gate()1 ? igid1 : IGId ? prim1 : Prim [IsReq (prim1)];
      gate ! igid1 ! prim1;
      (``''
        gate()1 ? igid2 : IGId ? prim2 : Prim ;
        gate ! igid2 ! prim2;
	exit
      []
        gate()2 ? igid3 : IGId ? prim3 : Prim [IsReq (prim3)];
        gate ! igid3 ! prim3;
	(``''
	  gate()1 ? igid4 : IGId ? prim4 : Prim;
	  exit
	|||
	  gate()2 ? igid5 : IGId ? prim5 : Prim;
	  exit
	)
      )
    endproc (* procname3 *)

  endproc (* procname1 *)
',
`    stop		(*** "proc_of($1)" is not unconfirmed ***)
  endproc (* procname1 *)
')'))procname1 [gate] (igid1, igid2)')

# "comb_file" gives the temporary filename used for combination processes

define(comb_file,maketemp(/tmp/combXXXXXX))

# "comb_procno" gives the next number to be used for a combination process

define(comb_procno,0)

# "disables(beh1,beh2)" produces a process definition and process
# instantiation for behaviour "beh1" to disable behaviour "beh2"

define(disables,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Disables`'comb_procno)file_save(comb_file,
`  '(* proc_of($1) may disable proc_of($2) *)
  
  process procname [gate] (igid1, igid2 : IGId) : exit :=
    $2
  [>
    $1
  endproc (* procname *)
)procname [gate] (igid1, igid2)')

# "enables(beh1,beh2)" produces a process definition and process instantiation
# for behaviour "beh1" to enable behaviour "beh2"

define(enables,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Enables`'comb_procno)file_save(comb_file,
`  '(* proc_of($1) will enable proc_of($2) *)
  
  process procname [gate] (igid1, igid2 : IGId) : exit :=
    $1
  >>
    $2
  endproc (* procname *)
)procname [gate] (igid1, igid2)')

# "enables_after_ack(facil,beh)" produces a process definition and process
# instantiation for acknowledging behaviour "facil" (which cannot be a
# combination) to enable "facil"; the following requires "facil" to be
# user_confirmed behaviour, but should take account of different patterns

define(enables_after_ack,
  `define(`comb_procno',
    incr(comb_procno))define(`procname1',
      EnablesAfterAck`'comb_procno)define(`procname2',
        procname1`'Aux)define(`procname3',
          procname1`'AuxA)define(`procname4',
            procname1`'AuxB)file_save(comb_file,
`  '(* acknowledging proc_of($1) will enable proc_of($2) *)
  
  process procname1 [gate] (igid1, igid2 : IGId) : noexit :=
ifelse(is_facil($1),0,
`    stop		(*** "proc_of($1)" is not one of the facilities ***)
  endproc (* procname1 *)
',
`ifelse(patt_of($1),user_confirmed,
`    hide gate()1``,'' gate()2 in
      (``''
	gate_qual(1,$1)
      |||
	gate_qual(2,$2)
      )
    |[gate()1``,'' gate()2]|
      procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    
  where
  
    process procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : noexit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      (``''
	[IsRes (prim1)] ->
	  (``''
	    [igid eq igid2] ->
	      procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
	  []
	    [igid eq igid1] ->
	      procname3 [gate``,'' gate()1``,'' gate()2] (igid2, igid1)
	  )
      []
        [not (IsRes (prim1))] ->
	  procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      )
    endproc (* procname2 *)
    
    process procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : exit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      (``''
        [IsCon (prim1)] ->
	  procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      []
        [not (IsCon (prim1))] ->
	procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      )
    []
      gate()2 ! igid2 ? prim2 : Prim [IsReq (prim2)];
      gate ! igid2 ! prim2;
      procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    endproc (* procname3 *)

    process procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : exit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    []
      gate()2 ? igid : IGId ? prim2 : Prim;
      gate ! igid ! prim2;
      procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    endproc (* procname4 *)

  endproc (* procname1 *)
',
`    stop		(*** "proc_of($1)" is not user_confirmed ***)
  endproc (* procname1 *)
')'))procname1 [gate] (igid1, igid2)')

# "enables_after_try(beh1,beh2)" produces a process definition and process
# instantiation for trying behaviour "beh1" to enable "beh2"

define(enables_after_try,
  `define(`comb_procno',
    incr(comb_procno))define(`procname1',
      EnablesAfterTry`'comb_procno)define(`procname2',
        procname1`'Aux)define(`procname3',
          procname1`'AuxA)define(`procname4',
            procname1`'AuxB)file_save(comb_file,
`  '(* trying proc_of($1) will enable proc_of($2) *)
  
  process procname1 [gate] (igid1, igid2 : IGId) : noexit :=
    hide gate()1``,'' gate()2 in
      (``''
	gate_qual(1,$1)
      |||
	gate_qual(2,$2)
      )
    |[gate()1``,'' gate()2]|
      procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    
  where
  
    process procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : noexit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      (``''
	[IsReq (prim1)] ->
	  (``''
	    [igid eq igid1] ->
	      procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
	  []
	    [igid eq igid2] ->
	      procname3 [gate``,'' gate()1``,'' gate()2] (igid2, igid1)
	  )
      []
        [not (IsReq (prim1))] ->
	  procname2 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      )
    endproc (* procname2 *)
    
    process procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : exit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      (``''
        [IsInd (prim1)] ->
	  procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      []
        [not (IsInd (prim1))] ->
	  procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
      )
    []
      gate()2 ! igid1 ? prim2 : Prim [IsReq (prim2)];
      gate ! igid1 ! prim2;
      procname3 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    endproc (* procname3 *)

    process procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2 : IGId) : exit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    []
      gate()2 ? igid : IGId ? prim2 : Prim;
      gate ! igid ! prim2;
      procname4 [gate``,'' gate()1``,'' gate()2] (igid1, igid2)
    endproc (* procname4 *)

  endproc (* procname1 *)
)procname1 [gate] (igid1, igid2)')

# "facil_attrs" gives the attributes of facilities as a list of triples:
# process name, pattern, property

define(facil_attrs,)

# "forall_ids(beh)" produces a process definition and process instantiation
# for behaviour "beh" for all igids

define(forall_ids,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Forall`'comb_procno)file_save(comb_file,
`  '(* proc_of($1) applies to all igids *)
  
  process procname [gate] : noexit :=
    choice igid1``,'' igid2 : IGId []
      [igid1 ne igid2] ->
        (
	  $1
	|||
	  procname [gate]
	)
  endproc (* procname *)
)procname [gate]')

# "interleaves(beh1,beh2)" produces a process instantiation and accumulated
# process definition for behaviour "beh1" to interleave with behaviour "beh2"

define(interleaves,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Interleaves`'comb_procno)file_save(comb_file,
`  '(* proc_of($1) is interleaved with proc_of($2) *)
  
  process procname [gate] (igid1, igid2 : IGId) : exit :=
    $1
  |||
    $2
  endproc (* procname *)
)procname [gate] (igid1, igid2)')

# "interrupts(beh1,beh2)" produces a process definition and process
# instantiation for behaviour "beh1" to interrupt behaviour "beh2" and then
# repeat

define(interrupts,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Interrupts`'comb_procno)file_save(comb_file,
`  '(* proc_of($1) may interrupt proc_of($2) then allow a repeat *)
  
  process procname [gate] (igid1, igid2 : IGId) : exit :=
    $2
  [>
    (``''
      $1
    >>
      procname [gate] (igid1, igid2)
    )
  endproc (* procname *)
)procname [gate] (igid1, igid2)')

# "interrupts_after_ack(beh1,beh2)" produces a process definition and process
# instantiation for behaviour "beh1" to interrupt after an acknowledgement of
# behaviour "beh2" and then allow a repeat

define(interrupts_after_ack,
  `define(`comb_procno',
    incr(comb_procno))define(`procname1',
      InterruptsAfterAck`'comb_procno)define(`procname2',
        procname1`'Aux)define(`procname3',
          procname1`'AuxA)file_save(comb_file,
`  '(* proc_of($1) may interrupt after an acknowledgement of proc_of($2), then repeat *)
  
  process procname1 [gate] (igid1, igid2 : IGId) : noexit :=
    hide gate()1 in
      gate_qual(1,$2)
    |[gate()1]|
      procname2 [gate``,'' gate()1] (igid1, igid2)
    
  where
  
    process procname2 [gate``,'' gate()1] (igid1, igid2 : IGId) : noexit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      (``''
	[IsRes (prim1)] ->
	  (``''
	    procname3 [gate``,'' gate()1] (igid1, igid2)
	  [>
	    (``''
	      $1
	    >>
	      procname1 [gate] (igid1, igid2)
	    )
	  )
      []
	[not (IsReq (prim1))] ->
	  procname2 [gate``,'' gate()1] (igid1, igid2)
      )
    endproc (* procname2 *)
    
    process procname3 [gate``,'' gate()1] (igid1, igid2 : IGId) : exit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      procname3 [gate``,'' gate()1] (igid1, igid2)
    endproc (* procname3 *)

  endproc (* procname1 *)
)procname1 [gate] (igid1, igid2)')

# "interrupts_after_try(beh1,beh2)" produces a process definition and process
# instantiation for behaviour "beh1" to interrupt after a try of behaviour
# "beh2" and then allow a repeat

define(interrupts_after_try,
  `define(`comb_procno',
    incr(comb_procno))define(`procname1',
      InterruptsAfterTry`'comb_procno)define(`procname2',
        procname1`'Aux)define(`procname3',
          procname1`'AuxA)file_save(comb_file,
`  '(* proc_of($1) may interrupt after a try of proc_of($2), then repeat *)
  
  process procname1 [gate] (igid1, igid2 : IGId) : noexit :=
    hide gate()1 in
      gate_qual(1,$2)
    |[gate()1]|
      procname2 [gate``,'' gate()1] (igid1, igid2)
    
  where
  
    process procname2 [gate``,'' gate()1] (igid1, igid2 : IGId) : noexit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      (``''
	[IsReq (prim1)] ->
	  (``''
	    procname3 [gate``,'' gate()1] (igid1, igid2)
	  [>
	    (``''
	      $1
	    >>
	      procname1 [gate] (igid1, igid2)
	    )
	  )
      []
	[not (IsReq (prim1))] ->
	  procname2 [gate``,'' gate()1] (igid1, igid2)
      )
    endproc (* procname2 *)
    
    process procname3 [gate``,'' gate()1] (igid1, igid2 : IGId) : exit :=
      gate()1 ? igid : IGId ? prim1 : Prim;
      gate ! igid ! prim1;
      procname3 [gate``,'' gate()1] (igid1, igid2)
    endproc (* procname3 *)

  endproc (* procname1 *)
)procname1 [gate] (igid1, igid2)')

# "overtakes(beh1,beh2)" produces a process definition and process
# instantiation for behaviour "beh1" (which must be a facility) to possibly
# overtake behaviour "beh2"

define(overtakes,
  `define(`comb_procno',
    incr(comb_procno))define(`procname1',
      Overtakes`'comb_procno)define(`procname2',
        procname1`'Aux)file_save(comb_file,
`  '(* proc_of($1) may possibly overtake proc_of($2) *)
  
  process procname1 [gate] (igid1, igid2 : IGId) : exit :=
ifelse(is_facil($1),0,
`    stop		(*** "proc_of($1)" is not one of the facilities ***)
  endproc (* procname2 *)',
  define(`group1',substr($1,0,ind))`    '(```'''
      $1
    |||
      $2
    )
  ||
    procname2 [gate] (igid1, igid2, <<>>)
    
  where
  
    process procname2 [gate]
     (igid1, igid2 : IGId, primq : PrimQueue) : exit :=
      gate ! igid1 ? prim1 : Prim;
      (```'''
        [IsReq (prim1)] ->
	  procname2 [gate] (igid1, igid2, prim1 + primq)
      []
        [not (IsReq (prim1))] ->
	  procname2 [gate] (igid1, igid2, primq)
      )
    []
      gate ! igid2 ? prim2 : Prim
       [IsInd (prim2) implies IsAheadQueue (prim2, primq)];
      (```'''
        [IsInd (prim2)] ->
	  procname2 [gate] (igid1, igid2, RemPrev (prim2, primq))
      []
        [not (IsInd (prim2))] ->
	  procname2 [gate] (igid1, igid2, primq)
      )
    endproc (* procname2 *)

  endproc (* procname1 *))
)procname1 [gate] (igid1, igid2)')

# "simultaneous(beh)" produces a process definition and process instantiation
# for behaviour "beh" interleaved in each direction

define(simultaneous,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Simultaneous`'comb_procno)file_save(comb_file,
`  '(* proc_of($1) is interleaved in both directions *)
  
  process procname [gate] (igid1, igid2 : IGId) : noexit :=
    $1
  |||
    igid_swap($1)
  endproc (* procname *)
)procname [gate] (igid1, igid2)')

# "unique_ids(group1,group2,beh)" produces a process definition and process
# instantiation for behaviour "beh" that ensures unique use of igids for
# service primitives "group1" that start associations and service primitives
# "group2" that end them; the following assumes unconfirmed behaviour for
# "group2", but should take account of different patterns

define(unique_ids,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Unique`'comb_procno)file_save(comb_file,
`  '(* in proc_of($3), igids are uniquely assigned/released by $1/$2 *)
  
  process procname [gate] (usedigids : IGIdSet) : noexit :=
    g ? igid : IGId ? prim : Prim
     [(IsGroup (prim, $1) and (IsReq (prim) or IsInd (prim))) implies
      (igid NotIn usedigids)];
    (``''
      [IsGroup (prim, $1)] ->
        procname [gate] (Insert (igid, usedigids))
    []
      [not (IsGroup (prim, $1) or IsGroup (prim, $2))] ->
        procname [gate] (usedigids)
    []
      [IsGroup (prim, $2)] ->
        procname [gate] (Remove (igid, usedigids))
    )
  endproc (* procname *)
)$3
||
  procname [gate] ({})')

# "withheld(group,beh)" produces a process definition and process
# instantiation for behaviour "beh" that may temporarily be refused

define(withheld,
  `define(`comb_procno',
    incr(comb_procno))define(`procname',
      Withheld`'comb_procno)file_save(comb_file,
`  '(* a $1 may temporarily not be allowed *)
  
  process procname [gate] : noexit :=
    choice igids : IGIdSet []	(* choose igid set allowing activity *)
      gate ? igid : IGId ? prim : Prim
       [(IsReq (prim) and IsGroup (prim, $1)) implies (igid IsIn igids)];
      procname [gate]
    []
      i;			(* revise igid set allowing activity *)
      procname [gate]
  endproc (* procname *)
)$2
||
  procname [gate]')

############################ Specification Macros ###########################

# "global(gate,beh)" produces all accumulated process and type definitions for
# gate "gate" and behaviour "beh"

define(global,
  `define(`procname',
    Service`'upper_of($1))define(`gate',
      lower_of($1))(*****************************************************************************

  Specification of upper_of($1) Service
  Generated by SAGE on syscmd(date)
*****************************************************************************)

specification procname [gate] : noexit
all_types(prims)
behaviour
  $2

where

sinclude(facil_file)sinclude(comb_file)endspec (* procname *)
file_delete(facil_file)file_delete(comb_file)')

########################### Trailer ##############################

divert
